import java.io.*;

/**
* A class to represent a block model in DH coordinate system
* @author Ho Ng
* @version 01/09/1998
*/
public class DHSystem
{
    private int totalFrames;    /* total number of block frames */

    private double[] dRX;
    private double[] dTX;
    private double[] dRZ;
    private double[] dTZ;

    private double[] rx;
    private double[] tx;
    private double[] rz;
    private double[] tz;

    private Matrix[] rtm;       /* transformation matrix of each frame */
    private int[] relative;


    /**
    * purpose : update the transformational matrix of a frame
    * @param frame frame number
    */
    private void updateRTM(int frame)
    {
        double sina = Math.sin(dRX[frame] + rx[frame]);
        double cosa = Math.cos(dRX[frame] + rx[frame]);
        double sint = Math.sin(dRZ[frame] + rz[frame]);
        double cost = Math.cos(dRZ[frame] + rz[frame]);

        rtm[frame].m[0][0] = cost;
        rtm[frame].m[0][1] = -sint;
        rtm[frame].m[0][3] = dTX[frame] + tx[frame];
        rtm[frame].m[1][0] = cosa * sint;
        rtm[frame].m[1][1] = cosa * cost;
        rtm[frame].m[1][2] = -sina;
        rtm[frame].m[1][3] = -sina * (dTZ[frame] + tz[frame]);
        rtm[frame].m[2][0] = sina * sint;
        rtm[frame].m[2][1] = sina * cost;
        rtm[frame].m[2][2] = cosa;
        rtm[frame].m[2][3] = cosa * (dTZ[frame] + tz[frame]);
    }

    /**
    * Constructor
    * purpose : initialize a DHSystem with number of parts
    * @param parts total number of parts or blocks
    */
    public DHSystem(int parts)
    {
        if (parts < 1)
            System.out.println("DHSystem : Failed to create object with no " +
                               "part");

        totalFrames = parts + 1;

        dRX = new double[totalFrames];
        dTX = new double[totalFrames];
        dRZ = new double[totalFrames];
        dTZ = new double[totalFrames];

        rx  = new double[totalFrames];
        tx  = new double[totalFrames];
        rz  = new double[totalFrames];
        tz  = new double[totalFrames];

        rtm = new Matrix[totalFrames];
        relative = new int[totalFrames];

        for (int i = 0; i < totalFrames; i++)
            rtm[i] = Matrix.identity();
    }

    /**
    * purpose : set all DH parameters of a frame
    * @param frame frame number
    * @param alpha rotation about x-axis
    * @param a translation along x-axis
    * @param theta rotation about z-axis
    * @param d translation along z-zxis
    * @param baseOn reference frame
    * @return true (succeed) or false (fail)
    */
    public boolean setDH(int frame, double alpha, double a, double theta,
                         double d, int baseOn)
    {
        if ((frame < 1) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.setDH : invalid frame number");
            return false;
        }

        dRX[frame] = alpha;
        dTX[frame] = a;
        dRZ[frame] = theta;
        dTZ[frame] = d;
        relative[frame] = baseOn;

        return reset(frame);
    }

    /**
    * purpose : reset all relative DH coordinates of a frame
    * @param frame frame number
    * @return true (succeed) or false (fail)
    */
    public boolean reset(int frame)
    {
        if ((frame < 1) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.reset : invalid frame number");
            return false;
        }

        rx[frame] = 0.0;
        tx[frame] = 0.0;
        rz[frame] = 0.0;
        tz[frame] = 0.0;
        updateRTM(frame);
        return true;
    }

    /**
    * purpose : get the relative angle alpha of a frame
    * @return relative angle alpha of the frame
    */
    public double getAlpha(int frame)
    {
        if ((frame < 1) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.getAlpha : invalid frame number");
            return 0.0;
        }

        return rx[frame];
    }

    /**
    * purpose : get the relative distance a of a frame
    * @return relative distance a of a frame
    */
    public double getA(int frame)
    {
        if ((frame < 1) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.getA : invalid frame number");
            return 0.0;
        }

        return tx[frame];
    }

    /**
    * purpose : get the relative angle theta of a frame
    * @return relative angle theta of a frame
    */
    public double getTheta(int frame)
    {
        if ((frame < 1) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.getTheta : invalid frame number");
            return 0.0;
        }

        return rz[frame];
    }

    /**
    * purpose : get the relative distance d of a frame
    * @return relative distance d of a frame
    */
    public double getD(int frame)
    {
        if ((frame < 1) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.getD : invalid frame number");
            return 0.0;
        }

        return tz[frame];
    }

    /**
    * purpose : change the relative angle alpha of a frame
    * @param frame frame number
    * @param angle new angle alpha
    * @return true (succeed) or false (fail)
    */
    public boolean changeAlpha(int frame, double angle)
    {
        if ((frame < 1) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.changeAlpha : invalid frame number");
            return false;
        }

        rx[frame] = angle;

        double sina = Math.sin(dRX[frame] + angle);
        double cosa = Math.cos(dRX[frame] + angle);
        double sint = Math.sin(dRZ[frame] + rz[frame]);
        double cost = Math.cos(dRZ[frame] + rz[frame]);

        rtm[frame].m[1][0] = cosa * sint;
        rtm[frame].m[1][1] = cosa * cost;
        rtm[frame].m[1][2] = -sina;
        rtm[frame].m[1][3] = -sina * (dTZ[frame] + tz[frame]);
        rtm[frame].m[2][0] = sina * sint;
        rtm[frame].m[2][1] = sina * cost;
        rtm[frame].m[2][2] = cosa;
        rtm[frame].m[2][3] = cosa * (dTZ[frame] + tz[frame]);

        return true;
    }

    /**
    * purpose : change the relative distance a of a frame
    * @param frame frame number
    * @param distance new distance a
    * @return true (succeed) or false (fail)
    */
    public boolean changeA(int frame, double distance)
    {
        if ((frame < 1) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.changeA : invalid frame number");
            return false;
        }

        tx[frame] = distance;
        rtm[frame].m[0][3] = dTX[frame] + distance;
        return true;
    }

    /**
    * purpose : change the relative angle theta of a frame
    * @param frame frame number
    * @param angle new angle theta
    * @return true (succeed) or false (fail)
    */
    public boolean changeTheta(int frame, double angle)
    {
        if ((frame < 1) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.changeTheta : invalid frame number");
            return false;
        }

        rz[frame] = angle;

        double sina = Math.sin(dRX[frame] + rx[frame]);
        double cosa = Math.cos(dRX[frame] + rx[frame]);
        double sint = Math.sin(dRZ[frame] + angle);
        double cost = Math.cos(dRZ[frame] + angle);

        rtm[frame].m[0][0] = cost;
        rtm[frame].m[0][1] = -sint;
        rtm[frame].m[1][0] = cosa * sint;
        rtm[frame].m[1][1] = cosa * cost;
        rtm[frame].m[2][0] = sina * sint;
        rtm[frame].m[2][1] = sina * cost;

        return true;
    }

    /**
    * purpose : change the relative distance d of a frame
    * @param frame frame number
    * @param distance new distance d
    * @return true (succeed) or false (fail)
    */
    public boolean changeD(int frame, double distance)
    {
        if ((frame < 1) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.changeD : invalid frame number");
            return false;
        }

        tz[frame] = distance;
        rtm[frame].m[1][3] = - Math.sin(dRX[frame] + rx[frame]) *
                             (dTZ[frame] + distance);
        rtm[frame].m[2][3] = Math.cos(dRX[frame] + rx[frame]) *
                             (dTZ[frame] + distance);
        return true;
    }

    /**
    * purpose : get the overall transformational matrix of a frame
    * @param frame frame number
    * @return overall transformational matrix of a frame
    */
    public Matrix getTransformation(int frame)
    {
        if ((frame < 0) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.getTransformation : invalid frame " +
                               "number");
            return Matrix.identity();
        }

        if (frame == 0)
            return Matrix.identity();

        Matrix result = new Matrix(rtm[frame]);
        int current = frame;

        while (relative[current] != 0)
        {
            current = relative[current];
            result = Matrix.multi(rtm[current], result);
        }

        return result;
    }

    /**
    * purpose : get the position of the origin of a frame
    * @param frame frame number
    * @return position of the origin of the frame
    */
    public Vector3D getFramePosition(int frame)
    {
        if ((frame < 0) || (frame >= totalFrames))
        {
            System.out.println("DHSystem.getFramePosition : invalid frame " +
                               "number");
            return (new Vector3D());
        }

        if (frame == 0)
            return (new Vector3D());

        Vector3D result = Vector3D.transformPointQ(rtm[frame], new Vector3D());
        int current = frame;

        while (relative[current] != 0)
        {
            current = relative[current];
            result = Vector3D.transformPointQ(rtm[current], result);
        }

        return result;
    }
};
